/*
 *	A Test of Lestrade: Asking Queries
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#include "ucw/lib.h"
#include "ucw/getopt.h"
#include "lestrade/lestrade.h"
#include "custom/custom.h"
#include "custom/debug.h"

#include <stdio.h>
#include <stdlib.h>

static void NONRET
usage(void)
{
  fputs("Usage: lt-query [-d <srcdir>] [-i <index>] <query>\n", stderr);
  exit(1);
}

int main(int argc, char **argv)
{
  byte *dir = "../data";
  byte *index = NULL;

  log_init(argv[0]);
  int opt;
  while ((opt = cf_getopt(argc, argv, CF_SHORT_OPTS "d:i:", CF_NO_LONG_OPTS, NULL)) >= 0)
    switch (opt)
      {
      case 'd':
	dir = optarg;
	break;
      case 'i':
	index = optarg;
	break;
      default:
	usage();
      }
  if (optind == argc)
    usage();

  lestrade_init(NULL);

  struct lestrade_query *Q = lestrade_new_query();
  Q->fetcher = test_card_fetch;
  Q->fetcher_priv = dir;
  Q->trace_mask = ~0U;

#if 0
  struct lestrade_expr *x = lestrade_new_match(Q);
  lestrade_add_atom(Q, x, "and")->sense = 1;
  lestrade_add_atom(Q, x, "more")->sense = 0;
  struct lestrade_expr *y = lestrade_new_match(Q);
  lestrade_add_atom(Q, y, "a-re-bourne");
  Q->expr = lestrade_new_op(Q, EX_OR, x, y);
#else
  struct lestrade_expr *x = lestrade_new_match(Q);
  while (optind < argc)
    {
      byte *c = argv[optind++];
      int sense = 0;
      if (*c == '+')
	{
	  sense = 1;
	  c++;
	}
      else if (*c == '-')
	{
	  sense = -1;
	  c++;
	}
      struct lestrade_atom *a = lestrade_add_atom(Q, x, c);
      a->sense = sense;
      a->accent_mode = ACCENT_MODE_STRIP;
      a->morph_mode = MORPH_MODE_PREACCENT;
    }
  Q->expr = x;
#endif

  if (lestrade_init_query(Q))
    {
      msg(L_ERROR, "Init error: %s (code %d)", Q->error_msg, Q->error_code);
      goto bed;
    }

  if (index)
    {
      msg(L_INFO, "Loading index %s", index);
      struct lestrade_index *idx;
      int err;
      if (err = lestrade_open_index(index, &idx))
	die("Cannot open index %s: error %d", index, err);
      lestrade_eval_index(Q, idx);
      lestrade_close_index(idx);
    }
  else
    {
      lestrade_eval_cards_init(Q);
      if (lestrade_eval_cards(Q, test_card_enum, dir) < 0)
	{
	  msg(L_ERROR, "Eval error: %s (code %d)", Q->error_msg, Q->error_code);
	  goto bed;
	}
      lestrade_eval_cards_cleanup(Q);
    }

  lestrade_get_results(Q);
  printf("Found %d results.\n", Q->num_results);
  for (uns i=0; i<Q->num_results; i++)
    printf("%3d. [%5d] %s\n", i+1, Q->results[i].quality, Q->results[i].attrs.name);

  if (Q->num_results)
    {
      printf("Printing result details:\n");
      lestrade_eval_detailed_init(Q);
      for (uns i=0; i<Q->num_results; i++)
	{
	  printf("### Result #%d\n", i+1);
	  struct lestrade_matches *m = lestrade_eval_detailed(Q, &Q->results[i].attrs);
	  for (uns j=0; j<m->num_matches; j++)
	    {
	      struct lestrade_match *t = &m->matches[j];
	      byte *s = m->text + m->words[t->pos].pos;
	      int len = m->words[t->pos+t->length-1].pos - m->words[t->pos].pos + m->words[t->pos+t->length-1].length;
	      len = MAX(len, 0);
	      printf("\t@%d <%.*s> w=%d $%d\n", t->pos, len, s, t->word_index, t->weight);
	    }
	}
      lestrade_eval_detailed_cleanup(Q);
    }

 bed:
  lestrade_free_query(Q);
  lestrade_cleanup();
  return 0;
}
