/*
 *	The Lestrade Search Library -- Matching Functions
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#define INFTY 1000000 			/* All Q's and weights are between -INFTY and INFTY */

struct trail_entry {
#ifdef CPU_LITTLE_ENDIAN		/* Trail entries are compared as u32's */
  s8 weight;
  byte word_id;
  u16 pos;
#else
  u16 pos;
  byte word_id;
  s8 weight;
#endif
};
COMPILE_ASSERT(trail_entry_size, sizeof(struct trail_entry) == 4);

void trail_enlarge(struct query *q);

static inline void
trail_init(struct query *q, uns sorted)
{
  q->trail_pos = q->trail;
  q->trail_sorted = sorted;
  q->trail_word_mask = 0;
  q->trail_bool_mask = 0;
}

static inline struct trail_entry *
trail_add_raw(struct query *q)
{
  if (unlikely(q->trail_pos >= q->trail_end))
    trail_enlarge(q);
  return q->trail_pos++;
}

static inline void
trail_add(struct query *q, uns pos, uns word_id, uns word_type, uns penalty)
{
  ASSERT(word_type < WT_MAX);
  struct word *w = q->words[word_id];
  if (!(w->type_set & (1 << word_type)))
    {
      DBG(">>> @%d w%d <%s> t%d ---", pos, word_id, q->words[word_id]->w, word_type);
      return;
    }
  if (pos >= 65536)
    {
      DBG(">>> @%d too far", pos);
      return;
    }

  int type_weight = word_type_weights[word_type];
  int wt = type_weight - penalty;
  DBG(">>> @%d w%d <%s> t%d $%d (%d-%d)", pos, word_id, q->words[word_id]->w, word_type, wt, type_weight, penalty);
  struct trail_entry *t = trail_add_raw(q);
  ASSERT(penalty < 256);
  t->pos = pos;
  t->word_id = word_id;
  t->weight = wt;
  if (!(q->trail_word_mask & (1 << word_id)))
    {
      q->trail_word_mask |= 1 << word_id;
      q->trail_bool_mask |= 1 << w->bool_id;
      w->best_weight = wt;
    }
  else
    w->best_weight = MIN(w->best_weight, wt);
}

int match_single(struct query *q, struct card_attrs *ca);

void matches_record_start(struct query *q);
void matches_record_stop(struct query *q);

/* match-full.c */

/* This is rather private, but match-detail is permitted to look there */
struct verbum {
  struct verbum *next;
  struct verbum *unaccented;
  byte class;
  byte num_occurs;
  byte word_idx[HARD_MAX_OCCURS];
  byte penalty[HARD_MAX_OCCURS];
  byte penalty_nonacc[HARD_MAX_OCCURS];
  byte has_accented_variants;
  byte word[0];				/* Must be aligned to a multiple of 4 bytes */
};
COMPILE_ASSERT(fh_verbum_align, !(OFFSETOF(struct verbum, word) % 4));

extern struct query *fh_query;
extern uns full_cnt;
struct verbum *fh_lookup(u16 *uni, uns ulen);
struct verbum *fh_lookup_nonacc(u16 *uni, uns ulen, uns *p_seen_accents);

int full_init(struct query *q);
int full_cleanup(struct query *q);
