/*
 *	The Lestrade Search Library -- Keeping of Results
 *
 *	(c) 2005 Martin Mares <mj@ucw.cz>
 */

#undef LOCAL_DEBUG

#include "ucw/lib.h"
#include "ucw/mempool.h"
#include "ucw/heap.h"
#include "lestrade/internal.h"

struct result_heap_entry {
  int quality;
  struct lestrade_result *note;
};

/* Result heap */

static inline int
rhe_lt(struct result_heap_entry *a, struct result_heap_entry *b)
{
  COMPARE_LT(a->quality, b->quality);
  return 0;
}
#define RHEAP_LESS(a,b) rhe_lt(&a,&b)
#define RHEAP_SWAP(h,a,b,t) (t=h[a], h[a]=h[b], h[b]=t)

void
results_init(struct query *q)
{
  q->result_max = q->Q->max_results;
  q->result_heap = mp_alloc(q->pool, q->result_max*sizeof(struct result_heap_entry));
  q->result_heap--;
  q->result_notes = mp_alloc(q->pool, q->result_max*sizeof(struct lestrade_result));
  q->result_cnt = 0;
}

void
results_add(struct query *q, struct card_attrs *ca, int qq, struct custom_result *cr)
{
  struct lestrade_result *n;
  if (q->result_cnt >= q->result_max)
    {
      if (q->result_heap[1].quality >= qq)
	{
	  custom_match_drop(q->Q, ca, cr);
	  return;
	}
      n = q->result_heap[1].note;
      custom_match_drop(q->Q, &n->attrs, &n->custom);
      DBG("RESULTS: dropped q=%d", q->result_heap[1].quality);
      HEAP_DELMIN(struct result_heap_entry, q->result_heap, q->result_cnt, RHEAP_LESS, RHEAP_SWAP);
    }
  else
    n = &q->result_notes[q->result_cnt];
  DBG("RESULTS: added q=%d", qq);
  n->quality = qq;
  n->attrs = *ca;
  n->custom = *cr;
  q->result_cnt++;
  q->result_heap[q->result_cnt].quality = qq;
  q->result_heap[q->result_cnt].note = n;
  HEAP_INSERT(struct result_heap_entry, q->result_heap, q->result_cnt, RHEAP_LESS, RHEAP_SWAP);
}

void
results_cleanup(struct query *q)
{
  for (uns i=1; i<=q->result_cnt; i++)
    custom_match_drop(q->Q, &q->result_notes[i].attrs, &q->result_notes[i].custom);
}

void
lestrade_get_results(struct lestrade_query *Q)
{
  struct query *q = Q->q;
  uns heap_cnt = q->result_cnt;
  for (uns i=1; i<=q->result_cnt; i++)
    HEAP_DELMIN(struct result_heap_entry, q->result_heap, heap_cnt, RHEAP_LESS, RHEAP_SWAP);

  DBG("Permuting results");
  struct result_heap_entry *heap = q->result_heap + 1;
  struct lestrade_result *notes = q->result_notes;
  for (uns i=0; i<q->result_cnt; i++)
    if (heap[i].note)
      {
	uns cycle_start = i;
	struct lestrade_result note0 = notes[i];
	DBG("Cycle starting at %d", cycle_start);
	do
	  {
	    uns j = heap[i].note - q->result_notes;
	    heap[i].note = NULL;
	    DBG("%d <- %d", i, j);
	    notes[i] = (j == cycle_start) ? note0 : notes[j];
	    i = j;
	  }
	while (i != cycle_start);
      }
  Q->results = notes;
  Q->num_results = q->result_cnt;
}
